"""Classes and methods for parsing log files.

Specifically logfiles generated by Gentoo Linux's emerge package manager.
"""

import re
from typing import Dict, List, Tuple

from .parser_package import PackageParser
from .report_objects import (
    DiskUsage,
    DiskUsageStats,
    LogInfo,
    PretendError,
    PretendSection,
    UpdateSection,
)


class Parser:
    """A class that provides methods to parse log files.

    Attributes
    ----------
    log_file (str): The name of the log file.
    log_data (List[str]): The content of the log file as a list of strings.
    """

    def __init__(self, log_file: str) -> None:
        """Initialize the Parser with a log file.

        Args:
        ----
            log_file: The name of the log file.
        """
        self.log_file = log_file
        self.log_data = self.read_log()

    def read_log(self) -> List[str]:
        """Read the log file and returns its content.

        Returns
        -------
            List[str]: The content of the log file as a list of strings.
        """
        with open(self.log_file, encoding="utf-8") as log_file:
            log_data = log_file.readlines()
        return self.split_log_to_sections(log_data)

    def split_log_to_sections(self, log_data: List[str]) -> Dict:
        """Split the log file into sections based on specified markers.

        Args:
        ----
            log_data (List[str]): Log content read with read_log method.

        Returns:
        -------
            Dict: A dictionary with section names as keys
                and content as values.
        """
        section_pattern = r"\{\{(.+?)\}\}"
        section_name = "beginning"
        log_by_sections: Dict[str, List[str]] = {section_name: []}

        for log_line in log_data:
            if " ::: " in log_line:
                line = log_line.split(" ::: ")[1].strip()
                match_section_pattern = re.search(section_pattern, line)
                if match_section_pattern:
                    section_name = "_".join(line.split()[1:-1]).lower()
                    log_by_sections[section_name] = []
                else:
                    log_by_sections[section_name].append(line)
            else:
                log_by_sections["final"] = log_line

        return log_by_sections

    def parse_emerge_pretend_section(
        self, section_content: List[str]
    ) -> PretendSection:
        """Parse the "emerge pretend" section of the log data.

        Args:
        ----
            section_content (List[str]): A list of strings that contains
                        the content of the "emerge pretend" section.

        Returns:
        -------
            Dict: A dictionary that contains the status of
                  the "emerge pretend" operation and the details.
        """
        if "emerge pretend was successful, updating..." in section_content:
            pretend_status = True
            pretend_details = None
        elif "There are no packages to update, skipping..." in section_content:
            pretend_status = True
            pretend_details = "No Updates"
        else:
            pretend_status = False
            pretend_details = self.parse_pretend_details(section_content)
        return PretendSection(pretend_status, pretend_details)

    def _parse_pretend_get_blocked_details(self, error_content: List[str]) -> List[str]:
        """Parse details of blocked package error.

        Args:
        ----
            error_content (List[str]): Lines from log marked ERROR by logger.

        Returns:
        -------
            List[str]: Parsed packages tht cause the error.
        """
        blocked_packages = []
        for line in error_content:
            match_package_pattern = re.search(r"^\((.+?)\)", line)
            if match_package_pattern:
                blocked_packages.append(match_package_pattern.group(1))

        return blocked_packages

    def _parse_pretend_get_error_type(
        self, error_content
    ) -> Tuple[str, str, List[str]]:
        """Parse error type and error details.

        Args:
        ----
            error_content (List[str]): Lines from log marked ERROR by logger.

        Returns:
        -------
            Tuple[str, str, List[str]]: Return error type, error definition as
              strings and error details as a list of strings.
        """
        error_type = "undefined"
        error_definition = "undefined"
        error_details = "undefined"

        for line in error_content:
            if "Blocked Packages" in line:
                error_type = "Blocked Packages"
                error_details = self._parse_pretend_get_blocked_details(error_content)

        for line in error_content:
            if line[0] == "*":
                error_definition += f"{line[2:]} "

        return (error_type, error_definition, error_details)

    def parse_pretend_details(self, section_content: List[str]) -> PretendError:
        """Parse information about the emerge pretend from logs.

        Args:
        ----
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
        -------
            PretendError: Dataclass that contains details about failed pretend,
                for example a list of blocked packages.
        """
        error_index = section_content.index("emerge pretend has failed, exiting")
        error_content = [line for line in section_content[error_index + 1 :] if line]

        error_type, _, error_details = self._parse_pretend_get_error_type(error_content)

        return PretendError(error_type, error_details)

    def parse_update_system_section(self, section_content: List[str]) -> UpdateSection:
        """Parse the "update system" section of the log data.

        Args:
        ----
            section_content (List[str]): A list of strings that contains
                    the content of the "update system" section.

        Returns:
        -------
            UpdateSection: A dataclass that contains the status
                  of the system update and package details.
        """
        try:
            update_type = (
                "@world" if section_content[2].split()[1] == "@world" else "security"
            )
        except IndexError:
            update_type = "Undefined"

        if "update was successful" in section_content:
            update_status = True
            if "Nothing to merge; quitting" in section_content:
                update_details = {"updated_packages": [], "errors": []}
            else:
                package_list = PackageParser().parse_update_details(section_content)
                update_details = {"updated_packages": package_list}
        elif "There are no packages to update, skipping..." in section_content:
            update_status = True
            update_type = "security"
            update_details = {"updated_packages": [], "errors": []}
        else:
            update_status = False
            update_details = {"updated_packages": [], "errors": []}
        return UpdateSection(update_type, update_status, update_details)

    def parse_disk_usage_info(self, section_content: List[str]) -> List[DiskUsageStats]:
        """Get disk usage information.

        Args:
        ----
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
        -------
            List[DiskUsageStats]: A list of DiskUsageStats objects
                for each mount point containing statistics of disk usage.
        """
        mount_point_lines = [
            line for line in section_content if line[0:14] == "Disk usage for"
        ]
        mount_points = []

        for line in mount_point_lines:
            split_content = line.split(" ===> ")
            stats = split_content[1].split(", ")

            mount_point = split_content[0].replace("Disk usage for ", "")
            total = stats[0].split("=")[1]
            used = stats[1].split("=")[1]
            free = stats[2].split("=")[1]
            percent_used = stats[3].split("=")[1]

            disk_usage_stats = DiskUsageStats(
                mount_point, total, used, free, percent_used
            )
            mount_points.append(disk_usage_stats)

        return mount_points

    def extract_info_for_report(self) -> LogInfo:
        """Extract information about the update from the log file.

        Returns
        -------
            LogInfo: Dataclass containing parsed data from all sections.
        """
        log_info = LogInfo
        disk_usage = DiskUsage
        for section, section_content in self.log_data.items():
            if section == "pretend_emerge":
                log_info.pretend_emerge = self.parse_emerge_pretend_section(
                    section_content
                )
            elif section == "update_system":
                log_info.update_system = self.parse_update_system_section(
                    section_content
                )
            elif section == "calculate_disk_usage_1":
                disk_usage.before_update = self.parse_disk_usage_info(section_content)
            elif section == "calculate_disk_usage_2":
                disk_usage.after_update = self.parse_disk_usage_info(section_content)
        log_info.disk_usage = disk_usage

        return log_info



"""Classes and methods for parsing log files.

Specifically logfiles generated by Gentoo Linux's emerge package manager.
"""

import re
from typing import List

from .report_objects import PackageInfo


class PackageParser:
    """A class that provides methods to parse package specific information."""

    def __init__(self) -> None:
        """Initialize PackageParser class."""
        pass

    def _parse_package_string(self, package_string: str) -> List[str]:
        """Parse package string into multiple parts.

        Args:
        ----
            package_string (str): String containing package information.

        Returns:
        -------
            List: List containing package information parts.
        """
        split_package_string = []
        temp = ""
        quotes_count = 0
        brackets_count = 0

        for char in package_string:
            temp += char
            if char == '"':
                quotes_count += 1
            elif char == "[":
                brackets_count += 1
            elif char == "]":
                brackets_count -= 1

            if char == " " and quotes_count % 2 == 0 and brackets_count == 0:
                split_package_string.append(temp.strip())
                temp = ""

        if temp:
            split_package_string.append(temp.strip())

        return split_package_string

    def _parse_package_ebuild(self, split_package_string: List) -> PackageInfo:
        """Parse ebuild information.

        Args:
        ----
            split_package_string (List[str]): A list with ebuild info, example:
                ['[ebuild     U  ]', 'sys-devel/gnuconfig-20230731::gentoo',
                 '[20230121::gentoo]', '72', 'KiB']

        Returns:
        -------
            PackageInfo: PackageInfo object with processed information
        """
        package_type = "ebuild"
        update_status = split_package_string[0]
        package_base_info = split_package_string[1]
        repo = package_base_info.split("::")[1]
        name_newversion = package_base_info.split("::")[0]

        package_name = ""
        for part in name_newversion.split("-"):
            if part.isnumeric() is True:
                pass
            elif "." in part or ":" in part:
                pass
            elif len(part) == 2 and part[0] == "r" and part[1].isnumeric():
                pass
            else:
                package_name += f"{part}-"

        new_version = name_newversion.replace(package_name, "")
        old_version = split_package_string[2].split("::")[0][1:]
        package_name = package_name[:-1]

        ebuild_info = PackageInfo(
            package_type,
            package_name,
            new_version,
            old_version,
            update_status,
            repo,
        )

        for var in split_package_string:
            if '="' in var:
                splitvar = var.split("=")
                ebuild_info.add_attributes({splitvar[0]: splitvar[1][1:-1].split(" ")})

        return ebuild_info

    def _parse_package_blocks(self, split_package_string: str) -> PackageInfo:
        """Parse blocks information.

        Args:
        ----
            split_package_string (List[str]): A list with blocks info, example:
                ['[blocks b      ]', '<perl-core/Compress-Raw-Zlib-2.204.1_rc',
                 '("<perl-core/Compress-Raw-Zlib-2.204.1_rc"', 'is', 'soft',
                 'blocking', 'virtual/perl-Compress-Raw-Zlib-2.204.1_rc)']

        Returns:
        -------
            PackageInfo: PackageInfo object with processed information
        """
        package_type = "blocks"
        package_name = split_package_string[1][1:]
        new_version = None
        old_version = None
        update_status = split_package_string[0]
        repo = None

        blocks_info = PackageInfo(
            package_type,
            package_name,
            new_version,
            old_version,
            update_status,
            repo,
        )

        blocks_info.add_attributes({"blocked_package": split_package_string[-1][:-1]})
        return blocks_info

    def _parse_package_uninstall(self, split_package_string: List) -> PackageInfo:
        """Parse uninstall information.

        Args:
        ----
            split_package_string (List[str]): A list with uninstall info

        Example:
        -------
            ['[uninstall     ]', 'perl-core/Compress-Raw-Zlib-2.202.0::gentoo']

        Returns:
        -------
            PackageInfo: PackageInfo object with processed information
        """
        package_type = "uninstall"
        split_package_info = split_package_string[1].split("::")
        package_name = split_package_info[0]
        repo = split_package_info[1]
        new_version = None
        old_version = None
        update_status = split_package_string[0]

        uninstall_info = PackageInfo(
            package_type,
            package_name,
            new_version,
            old_version,
            update_status,
            repo,
        )

        uninstall_info.add_attributes({"uninstalled_package": package_name})
        return uninstall_info

    def parse_update_details(self, section_content: List[str]) -> List[PackageInfo]:
        """Parse information about update from log file.

        Args:
        ----
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
        -------
            List[PackageInfo]: List of PackageInfo objects where each object
                contains useful information for the report.
        """
        ebuild_info_pattern = r"\[(.+?)\]"
        package_strings = [
            line
            for line in section_content
            if re.search(ebuild_info_pattern, line) and line != "[ ok ]"
        ]
        packages = []
        for package_string in package_strings:
            split_package_string = self._parse_package_string(package_string)
            update_status = split_package_string[0]

            if "ebuild" in update_status:
                package = self._parse_package_ebuild(split_package_string)
            elif "blocks" in update_status:
                package = self._parse_package_blocks(split_package_string)
            elif "uninstall" in update_status:
                package = self._parse_package_uninstall(split_package_string)

            packages.append(package)

        return packages



"""Dataclasses that are used to construct update reports."""

from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass
class PackageInfo:
    """Dataclass that contains information about a package."""

    package_type: str
    package_name: str
    new_version: str
    old_version: str
    update_status: str
    repo: str

    def add_attributes(self, attrs):
        """Add attributes to the PackageInfo object."""
        for attr_name, attr_value in attrs.items():
            setattr(self, attr_name, attr_value)


@dataclass
class UpdateSection:
    """Dataclass update section."""

    update_type: str
    update_status: bool
    update_details: Dict[str, PackageInfo]


@dataclass
class PretendError:
    """Dataclass pretend error."""

    error_type: str
    error_details: List[str]


@dataclass
class PretendSection:
    """Dataclass pretend section."""

    pretend_status: bool
    pretend_details: PretendError


@dataclass
class DiskUsageStats:
    """Dataclass disk usage stats."""

    mount_point: str
    total: str
    used: str
    free: str
    percent_used: str


@dataclass
class DiskUsage:
    """Dataclass disk usage."""

    before_update: List[DiskUsageStats]
    after_update: List[DiskUsageStats]


@dataclass
class LogInfo:
    """Dataclass log info."""

    pretend_emerge: Optional[PretendSection]
    update_system: Optional[UpdateSection]
    disk_usage: Optional[DiskUsage]



"""Automate updates on Gentoo Linux.

This module provides a command-line interface for automating updates on
Gentoo Linux.
It defines CLI command flags using argparse and provides functions for
reading make.conf, creating log directories, and generating reports.
"""

import argparse
import os
import sys
from typing import Dict, List, Tuple

from ._version import __version__
from .notifier import Notifier
from .parser import Parser
from .reporter import Reporter
from .shell_runner import ShellRunner

current_path = os.path.dirname(os.path.realpath(__file__))
sys.tracebacklimit = -1


def create_cli() -> argparse.Namespace:
    """Define CLI command flags using argparse.

    Returns
    -------
        parser.parse_args: Argparse commands
    """
    formatter = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(
        description="Automate updates on Gentoo Linux.",
        formatter_class=formatter,
    )

    subparsers = parser.add_subparsers(dest="command")
    update = subparsers.add_parser("update", help="Run security or full update.")
    report = subparsers.add_parser("report", help="Generate or send update reports.")
    version = subparsers.add_parser("version", help="Print gentoo-update version.")

    # define update subparser
    update.add_argument(
        "-m",
        "--update-mode",
        default="security",
        choices=["security", "full"],
        help="""
Set the update mode.
Options:
* security: update only security patches (GLSA)
* full: do a full @world update
Default: security
""",
    )
    update.add_argument(
        "-a",
        "--args",
        default="",
        help="""
Additional arguments to be passed when in 'full' update mode.
Example:
--args 'quiet-build=n color=y keep-going'
""",
    )
    update.add_argument(
        "-c",
        "--config-update-mode",
        default="ignore",
        choices=["ignore", "merge"],
        help="""
Set the way new configurations are handled after an update.
Options:
* ignore: do not update configuration files at all.
* merge: automatically merge changes in configuration files.
Default: ignore
""",
    )
    update.add_argument(
        "-u",
        "--disk-usage-limit",
        default="0",
        help="""
Do not run update if available disk space is lower than a limit (in GB).
Default: 0 - do not set a limit.
""",
    )
    update.add_argument(
        "-d",
        "--daemon-restart",
        action="store_true",
        help="Set whether to restart services and daemons after an update.",
    )
    update.add_argument(
        "-e",
        "--clean",
        action="store_true",
        help="Set whether to clean orphaned packaged after an update.",
    )
    update.add_argument(
        "-l",
        "--read-logs",
        action="store_true",
        help="Set whether to read elogs after an update.",
    )
    update.add_argument(
        "-n",
        "--read-news",
        action="store_true",
        help="Set whether to read news after an update.",
    )
    update.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Do not show logs on the terminal screen.",
    )

    # define report subparser
    report.add_argument(
        "-r",
        "--report",
        nargs="?",
        const="LAST",
        default=None,
        help="Show report. By default shows report from the last update log.",
    )
    report.add_argument(
        "-o",
        "--last-n-logs",
        type=int,
        help="Show last n log filenames.",
    )
    report.add_argument(
        "-s",
        "--send-report",
        default="none",
        choices=["irc", "email", "mobile", "none"],
        help="""
Send update report via IRC bot, email (SendGrid) or mobile app.
Default: none
""",
    )
    report.add_argument(
        "-t",
        "--short-report",
        action="store_true",
        help="Show or send only update status without package info.",
    )

    # define version subparser
    version.add_argument("-v", "--verbose", action="store_true")

    args = parser.parse_args()
    return args


def make_conf_reader() -> Dict:
    """Read /etc/portage/make.conf.

    Returns
    -------
        Dict: Parameters in the key:value format.
        Example: {'COMMON_FLAGS': '"-O2 -pipe"'}
    """
    make_conf = {}
    with open("/etc/portage/make.conf", encoding="utf-8") as make_conf_raw:
        lines = make_conf_raw.read().splitlines()
    key, value = None, []
    for line in lines:
        line_stripped = line.strip()
        if not line_stripped or line_stripped.startswith("#"):
            continue
        if "=" in line_stripped:
            if key:
                make_conf[key] = " ".join(value)
            parts = line_stripped.split("=", 1)
            key = parts[0].strip()
            value = [parts[1].strip()]
        else:
            value.append(line_stripped)
    if key:
        make_conf[key] = " ".join(value)
    return make_conf


def initiate_log_directory(make_conf) -> Tuple[str, List[str]]:
    """Create log directory if it does not exist.

    If PORTAGE_LOGDIR is not set, use the default directory.

    Returns
    -------
        str: Log directory path.
        List[str]: List of messages to be logged.
    """
    try:
        log_dir = make_conf["PORTAGE_LOGDIR"].replace('"', "")
    except KeyError:
        log_dir = ""

    log_dir_messages = []
    if log_dir == "":
        log_dir = "/var/log/portage/gentoo-update"
        log_dir_messages.append(f"PORTAGE_LOGDIR not set, using default: {log_dir}")
    else:
        log_dir = log_dir.replace('"', "")
        log_dir = f"{log_dir}/gentoo-update"
        log_dir_messages.append(f"PORTAGE_LOGDIR set, using: {log_dir}")

    if not os.path.exists(log_dir):
        os.mkdir(log_dir)
        log_dir_messages.append(f"Created log directory: {log_dir}")

    return log_dir, log_dir_messages


def show_available_reports(log_dir: str, last_n_logs: int) -> None:
    """Short last n reports in the log directory.

    Args
    ----
        log_dir (str): Directory where gentoo_update stores logs.
        last_n_logs (int): Last n amount of reports in the directory.
    """
    log_filesnames = os.listdir(log_dir)
    log_filesnames.sort()
    if len(log_filesnames) < last_n_logs:
        raise ValueError(f"There are less than {last_n_logs} in {log_dir}")

    if not log_filesnames:
        raise ValueError(f"No log files found in the directory {log_dir}")

    logs = log_filesnames[-last_n_logs:]
    print(f"The last {last_n_logs} log file filenames")
    for log in logs:
        print(log)


def generate_report(
    log_dir: str, log_filename: str = None, short_report: bool = False
) -> Reporter:
    """Show report for the <log_filename> located in $PORTAGE_LOGDIR.

    Args
    ----
        log_dir (str): Directory where gentoo_update stores logs.
        log_filename (str, optional): File name of the update log.
                Defaults to the latest log if not provided.
        short_report (bool): Short report format.

    Returns
    -------
        Reporter: Object containing the update report information.
    """
    if not os.path.exists(log_dir):
        raise FileNotFoundError(f"The log directory {log_dir} does not exist")

    if log_filename is None:
        files = os.listdir(log_dir)
        paths = [
            os.path.join(log_dir, basename)
            for basename in files
            if basename[0:4] == "log_"
        ]
        if not paths:
            raise ValueError(f"No log files found in the directory {log_dir}")
        full_log_path = max(paths, key=os.path.getctime)
    else:
        full_log_path = os.path.join(log_dir, log_filename)

        if not os.path.exists(full_log_path):
            raise FileNotFoundError(
                f"The log file {log_filename} does not exist in {log_dir}"
            )

    update_info = Parser(full_log_path).extract_info_for_report()
    return Reporter(update_info, short_report)


def main() -> None:
    """Execute it all."""
    args = create_cli()

    make_conf = make_conf_reader()
    log_dir, log_dir_messages = initiate_log_directory(make_conf)

    if args.command == "version":
        if args.verbose:
            print(f"gento-update version: {__version__}")
        else:
            print(__version__)
    elif args.command == "update":
        runner = ShellRunner("y" if args.quiet else "n", log_dir, log_dir_messages)
        runner.run_shell_script(
            args.update_mode,
            args.args if args.args else "NOARGS",
            args.disk_usage_limit,
            args.config_update_mode,
            "y" if args.daemon_restart else "n",
            "y" if args.clean else "n",
            "y" if args.read_logs else "n",
            "y" if args.read_news else "n",
        )
    elif args.command == "report":
        if args.last_n_logs:
            show_available_reports(log_dir, args.last_n_logs)
        elif args.send_report in ["irc", "email", "mobile"]:
            log_filename = None if args.report == "LAST" else args.report
            report = generate_report(
                log_dir, log_filename, args.short_report
            ).create_report()
            short = False if args.send_report != "irc" else True
            Notifier(notification_type=args.send_report, report=report, short=short)
        else:
            log_filename = None if args.report == "LAST" else args.report
            report = generate_report(log_dir, log_filename, args.short_report)
            report.print_report()


if __name__ == "__main__":
    main()



"""Unit tests for reporter.py file."""

import unittest

from gentoo_update import Reporter, generate_last_report


class TestGentooUpdate(unittest.TestCase):
    """Unit tests for the gentoo_update module."""

    def setUp(self):
        """Initialize test prerequisites."""
        report_for_tests = "logs_for_unit_tests"
        self.report_object = generate_last_report(report_for_tests, short_report=False)
        self.report = self.report_object.create_report()

    def test_report_object_type(self):
        """Test if the report_object is of type Reporter."""
        self.assertIs(type(self.report_object), Reporter)

    def test_report_status(self):
        """Test if the report status is SUCCESS."""
        status = self.report[1].split(" ")[-1]
        self.assertEqual(status, "SUCCESS")

    def test_report_packages(self):
        """Test if the report lists the expected packages."""
        packages = sum([pkg.split(" ", 1)[1:] for pkg in self.report[3:7]], [])
        correct_package_list = [
            "dev-libs/openssl 3.0.10:0/3->3.0.11:0/3",
            "net-fs/samba 4.18.4-r1->4.18.8",
            "net-libs/nghttp2 1.51.0:0/1.14->1.57.0:0/1.14",
            "www-client/firefox-bin 118.0.1:rapid->118.0.2:rapid",
        ]
        self.assertEqual(packages, correct_package_list)

    def test_disk_usage(self):
        """Test if the report shows the expected disk usage."""
        disk_usage = self.report[-4::]
        correct_disk_usage = [
            "Mount Point /",
            "Free Space 253G => 253G",
            "Used Space 177G => 177G",
            "Used (%) 42% => 42%",
        ]
        self.assertEqual(disk_usage, correct_disk_usage)


if __name__ == "__main__":
    unittest.main()



"""Provides a class `ShellRunner` for running shell scripts."""

import logging
import os
import subprocess
import sys
from datetime import datetime
from typing import List


class ShellRunner:
    """Runs shell scripts and logs the output to a file and terminal.

    Args:
    ----
        quiet (str): If 'y', suppresses terminal output.
        log_dir (str): Directory to save log files.
        log_dir_messages (str): List of messages to log.

    Attributes:
    ----------
        quiet (bool): If True, suppresses terminal output.
        timestamp (str): Current timestamp in the format of '%Y-%m-%d-%H-%M'.
        log_dir (str): Directory to save log files.
        log_dir_messages (str): List of messages to log.
        log_filename (str): Log filename.
        logger (logging.Logger): Configured logger.
        script_dir (str): Directory of the shell script.
        script_path (str): Path to the shell script.
        stdout_output (List[str]): List containing the standard output.
        stderr_output (List[str]): List containing the standard error output.
    """

    def __init__(self, quiet: str, log_dir: str, log_dir_messages: str) -> None:
        """Initialize ShellRunner class."""
        self.quiet = True if quiet == "y" else False

        self.timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M")
        self.log_dir = log_dir
        self.log_dir_messages = log_dir_messages

        self.log_filename = f"{self.log_dir}/log_{self.timestamp}"
        self.logger = self.initiate_logger()

        self.script_dir = os.path.join(os.path.dirname(__file__), "scripts")
        self.script_path = os.path.join(self.script_dir, "updater.sh")

    def initiate_logger(self) -> logging.Logger:
        """Create a logger with two handlers.

            1. terminal output
            2. file output

        Both handlers have the same logging level (INFO)
        and share the same formatter.
        Formatters include timestamp, log level and the message.

        Returns
        -------
            logging.Logger: Configured logger.
            log_filename: Log filename.
        """
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)

        formater = logging.Formatter(
            "[%(asctime)s %(levelname)s] ::: %(message)s",
            datefmt="%d-%b-%y %H:%M:%S",
        )

        if not self.quiet:
            terminal_handler = logging.StreamHandler()
            terminal_handler.setLevel(logging.INFO)
            terminal_handler.setFormatter(formater)
            logger.addHandler(terminal_handler)

        try:
            file_handler = logging.FileHandler(self.log_filename)
        except PermissionError:
            print("Not enough permissions to create a log file, exiting")
            sys.exit(1)
        file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(formater)
        logger.addHandler(file_handler)

        for message in self.log_dir_messages:
            logger.info(message)

        return logger

    def _log_stream_output(
        self, stream_obj: subprocess.Popen, outputtype: str
    ) -> List[str]:
        """Process sterr from the upadte script.

        Args:
        ----
            stream_obj (subprocess.Popen): output stream from subprocess.Popen
            outputtype (str): output type, stderr or stdout

        Returns:
        -------
            List[str]: List containing the output
        """
        output = []
        if outputtype == "stdout":
            stream = stream_obj.stdout
        elif outputtype == "stderr":
            stream = stream_obj.stderr
        else:
            self.logger.error("Invalid Output Stream Type")

        for stream_line in stream:
            line = stream_line.decode().rstrip("\n")
            output.append(line)
            if outputtype == "stdout":
                self.logger.info(line)
            elif outputtype == "stderr":
                self.logger.error(line)
        return output

    def _exit_with_error_message(self, stream: subprocess.Popen) -> None:
        """Exit runner if updater.sh encounters an error and log that error.

        Args:
        ----
            stream: output stream from subprocess.Popen
            stderr_output: List containing the standard error output.
        """
        error_message = "updater.sh exited with error code {script_stream.returncode}"
        if self.stderr_output:
            stderr_output_message = "n".join(self.stderr_output)
            error_message += f"\nStandard error output:\n{stderr_output_message}"
        self.logger.error(error_message)
        sys.exit(stream.returncode)

    def run_shell_function(self, command: List) -> None:
        """Run a shell script and stream standard output.

        and standard error to terminal and a log file.

        Args:
        ----
            command (List(str)): A call to specific function in
                                 update.sh with all parameters.
        """
        with subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ) as script_stream:
            self.stdout_output = self._log_stream_output(script_stream, "stdout")
            self.stderr_output = self._log_stream_output(script_stream, "stderr")
            script_stream.wait()

            if script_stream.returncode != 0:
                self._exit_with_error_message(script_stream)

    def run_shell_script(self, *args: str) -> None:
        """Run every function in update.sh one by one.

        Args:
        ----
            *args (str): Arguments for the shell script.
                         They need to be handled by the script.
        """
        script_stages = [
            "check_root_part_limit",
            "check_disk_usage_before_update",
            "sync_tree",
            "emerge_pretend",
            "update",
            "config_update",
            "clean_up",
            "check_restart",
            "get_logs",
            "get_news",
            "check_disk_usage_after_update",
        ]
        for stage in script_stages:
            command = [self.script_path] + [stage] + list(args)
            self.run_shell_function(command)

        final_message = f"gentoo-update is done! Log:file: {self.log_filename}"
        self.logger.info(final_message)
        if self.quiet:
            print(final_message)

    def __del__(self) -> None:
        """Close all file handlers after ShellRunner is closed."""
        try:
            if self.logger:
                for handler in self.logger.handlers:
                    try:
                        handler.close()
                        self.logger.removeHandler(handler)
                    except Exception as exc:
                        print(f"Handler Error: Could not close handler: {exc}")
        except AttributeError as attrerr:
            print(f"Handler Error: {attrerr}. Logger object was not defined")



"""Provides functionality for updating a Gentoo Linux system."""

from ._version import __version__
from .gentoo_update import create_cli, generate_report, main
from .notifier import Notifier
from .parser import Parser
from .parser_package import PackageParser
from .report_objects import (
    DiskUsage,
    DiskUsageStats,
    LogInfo,
    PackageInfo,
    PretendError,
    PretendSection,
    UpdateSection,
)
from .reporter import Reporter
from .shell_runner import ShellRunner



"""Contains the Reporter class which is responsible for generating a report.

The report can be either a successful or failed update report.
The class has methods for creating a report when emerge pretend fails,
creating a report when update fails, creating a report when update succeeds,
and creating a report.

The module also contains helper classes for parsing log information.
"""
import sys
from typing import List

from .parser import DiskUsage, LogInfo, PretendError, PretendSection, UpdateSection


class Reporter:
    """Reporter class for generating a report."""

    def __init__(self, update_info: LogInfo, short_report: bool) -> None:
        """Initialize Reporter class."""
        self.info = update_info
        self.short_report = short_report

    def _create_failed_pretend_report(self, pretend_info: PretendSection) -> List[str]:
        """Create a report when emerge pretend fails.

        Args:
        ----
            pretend_info (PretendSection): emerge pretend information.

        Returns:
        -------
            List: A list of strings that comprise the failed pretend report.
        """
        report = [
            "==========> Gentoo Update Report <==========",
            "emerge pretend status: FAIL",
        ]
        pretend_details = pretend_info.pretend_details
        if pretend_details.error_type == "undefined":
            report.append("Could not identify error, please check the logs")
        elif pretend_details.error_type == "Blocked Packages":
            report.append(f"\nError Type: {pretend_details.error_type}")
            report.extend(self._report_blocked_packages(pretend_details))
        report.append("")
        return report

    def _report_blocked_packages(self, pretend_details: PretendError) -> List[str]:
        """Report on Blocked Packages error during emerge pretend.

        Args:
        ----
            pretend_details (PretendError): In this case it's blocked packages.

        Returns:
        -------
            List[str]: Section of the report about blocked packages.
        """
        blocked_packages_report = ["List of Blocked Packages:"]
        for package in pretend_details.error_details:
            blocked_packages_report.append(f"-----> {package}")

        return blocked_packages_report

    def _create_failed_report(
        self, update_info: UpdateSection, disk_usage_info: DiskUsage
    ) -> List[str]:
        """Create a report when update fails.

        Args:
        ----
            update_info (LogInfo.UpdateSection): Update information.
            disk_usage_info (LogInfo.DiskUsage): Disk usage information.

        Returns:
        -------
            List: A list of strings that comprise the failed update report.
        """
        # do failed report processing
        return ["Your update is a failure"]

    def _create_successful_report(
        self, update_info: UpdateSection, disk_usage_info: DiskUsage
    ) -> List[str]:
        """Create a report when update succeeds.

        Args:
        ----
            update_info (LogInfo.UpdateSection): Update information.
            disk_usage_info (LogInfo.DiskUsage): Disk usage information.

        Returns:
        -------
            List: A list of strings that comprise the successful update report.
        """
        report = [
            "==========> Gentoo Update Report <==========",
            "update status: SUCCESS",
        ]
        updated_packages = update_info.update_details["updated_packages"]
        other_package_types = False
        if updated_packages:
            report.append(r"processed packages:")
            for package in updated_packages:
                if package.package_type == "ebuild":
                    package_name = package.package_name
                    new_version = package.new_version
                    old_version = package.old_version
                    report.append(f"--- {package_name} {old_version}->{new_version}")
                else:
                    other_package_types = True

            if other_package_types:
                report.append("")
                report.append("Non-ebuild packages")
                for package in updated_packages:
                    if package.package_type == "blocks":
                        package_name = package.package_name
                        blocked_package = package.blocked_package
                        report.append(f"--- {package_name} blocked {blocked_package}")
                    elif package.package_type == "uninstall":
                        package_name = package.package_name
                        # uninstalled_package seems unused
                        # uninstalled_package = package.uninstalled_package
                        report.append(f"--- {package_name} was uninstalled")
            report.append("")
            report.append("Disk Usage Stats:")

            for before, after in zip(
                disk_usage_info.before_update, disk_usage_info.after_update
            ):
                disk_usage_stats = [
                    f"Mount Point {before.mount_point}",
                    f"Free Space {before.free} => {after.free}",
                    f"Used Space {before.used} => {after.used}",
                    f"Used (%) {before.percent_used} => {after.percent_used}",
                ]
                report += disk_usage_stats

        return report

    def create_report(self) -> List[str]:
        """Create a report.

        Returns
        -------
            List: A list of strings that comprise the update report.
        """
        info = self.info
        try:
            disk_usage_info = info.disk_usage
            pretend_success = info.pretend_emerge.pretend_status

            if pretend_success:
                update_info = info.update_system
                update_success = info.update_system.update_status
            else:
                pretend_info = info.pretend_emerge
                update_success = False

            if self.short_report:
                update_status = "SUCCESS" if update_success else "FAIL"
                return [f"update status: {update_status}"]

            if update_success:
                report = self._create_successful_report(update_info, disk_usage_info)
            elif pretend_success and not update_success:
                report = self._create_failed_report(update_info, disk_usage_info)
            else:
                report = self._create_failed_pretend_report(pretend_info)
            return report
        except AttributeError:
            print("[Error] Could not create the report, incomplete log file")
            sys.exit(1)

    def print_report(self) -> None:
        """Print the report line by line to console."""
        report = self.create_report()
        for line in report:
            print(line)



"""Provides a Notifier to send update reports via email, IRC or mobile."""

import json
import os
import socket
import ssl
import time
import urllib.request
from sys import exit
from typing import List, Tuple

ACCEPTED_HTTP_CODES = [200, 202]
USE_SENDGRID = True
try:
    import sendgrid  # noqa: I005
    from sendgrid.helpers.mail import Content, Email, Mail, To  # noqa: I005
except ImportError:
    USE_SENDGRID = False


class Notifier:
    """Notifier class for sending update reports."""

    def __init__(self, notification_type: str, report: List, short=True) -> None:
        """Initialize Notifier class."""
        report = report[0:2] if short else report

        if notification_type == "email":
            if USE_SENDGRID:
                self.send_report_to_mail(report)
            else:
                print("sendgrid library is not installed")
                print("it can be installed from GURU overlay:")
                print("  emerge --ask dev-python/sendgrid")
        elif notification_type == "irc":
            self.send_report_to_irc(report)
        elif notification_type == "mobile":
            self.send_report_to_mobile(report)
        else:
            print("Unsupported authentication methods")
            print("Currently supporting: irc")
            print("Exiting...")

    def get_irc_vars(self) -> Tuple[str, str, str]:
        """Get variables needed to send report to IRC chat from env."""
        channel = os.getenv("IRC_CHANNEL")
        botnick = os.getenv("IRC_BOT_NICKNAME")
        botpass = os.getenv("IRC_BOT_PASSWORD")
        if None not in (channel, botnick, botpass):
            return channel, botnick, botpass
        else:
            print("Undefined enviromental variable(s)")
            print("Define: IRC_CHANNEL, IRC_BOT_NICKNAME, IRC_BOT_PASSWORD")
            exit(1)

    def send_report_to_irc(self, report: List[str]) -> None:
        """Send the update report to IRC chat."""
        server = "irc.libera.chat"
        port = 6697
        channel, botnick, botpass = self.get_irc_vars()
        ssl_context = ssl.create_default_context()

        irc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        irc = ssl_context.wrap_socket(irc, server_hostname=server)
        irc.connect((server, port))
        irc.send(f"USER {botnick} {botnick} {botnick} {botnick}\n".encode())
        irc.send(f"NICK {botnick}\n".encode())
        irc.send(f"PRIVMSG NickServ :IDENTIFY {botnick} {botpass}\n".encode())
        time.sleep(5)

        irc.send(f"JOIN {channel}\n".encode())
        for line in report:
            irc.send(f"PRIVMSG {channel} :{line}\n".encode())
            time.sleep(10)
        print("report sent, quitting...")

        irc.send(b"QUIT \n")
        irc.close()

    def get_mail_vars(self) -> Tuple[str, str, str]:
        """Get variables to send report to email via SendGrid from env."""
        api_key = os.getenv("SENDGRID_API_KEY")
        send_to = os.getenv("SENDGRID_TO")
        send_from = os.getenv("SENDGRID_FROM")
        if None not in (api_key, send_to, send_from):
            return api_key, send_to, send_from
        else:
            print("Undefined enviromental variable(s)")
            print("Please define: SENDGRID_API_KEY, SENDGRID_TO, SENDGRID_FROM")
            exit(1)

    def send_report_to_mail(self, report: List[str]) -> None:
        """Send the update report to email via SendGrid."""
        api_key, send_to, send_from = self.get_mail_vars()
        sendgrid_client = sendgrid.SendGridAPIClient(api_key=api_key)
        subject = "Gentoo Linux Update Report"

        content = Content("text/plain", "\n".join(report))
        mail = Mail(Email(send_from), To(send_to), subject, content)
        mail_json = mail.get()

        response = sendgrid_client.client.mail.send.post(request_body=mail_json)
        if response.status_code in ACCEPTED_HTTP_CODES:
            print("email was sent successfully!")
        else:
            print("email was not sent successfully, details:")
            print(response.headers)
            print(response.body)

    def send_report_to_mobile(self, report: List[str]) -> None:
        """Send the update report to mobile app."""
        token = os.getenv("GU_TOKEN")
        if not token:
            print("Token not found, please define GU_TOKEN env variable.")
            exit(1)
        update_status = report[1].split(": ")[1]
        update_content = report[2:]

        url = "https://us-central1-gentoo-update.cloudfunctions.net/checkTokenAndForwardData"
        headers = {"Content-Type": "application/json"}
        data = {
            "token": token,
            "update_status": update_status,
            "update_content": update_content,
        }

        data = json.dumps(data).encode("utf-8")

        req = urllib.request.Request(url, data=data, headers=headers, method="POST")
        with urllib.request.urlopen(req) as response:
            print(response.status)
            print(response.read().decode("utf-8"))


if __name__ == "__main__":
    report = [
        "==========> Gentoo Update Report <==========",
        "update status: SUCCESS",
        "processed packages:",
        "--- sys-apps/sandbox 2.32->2.37",
        "",
        "Disk Usage Stats:",
        "Free Space 35G => 35G",
        "Used Space 7.5G => 7.5G",
        "Used pc(%) 18% => 18%",
    ]
    notify = Notifier(notification_type="mobile", report=report, short=False)



"""Unit tests for gentoo_update.py (main) file."""

import unittest
from unittest.mock import Mock, patch

from gentoo_update import create_cli


class TestGentooUpdate(unittest.TestCase):
    """Unit tests for functions in the gentoo_update module."""

    @patch("argparse.ArgumentParser.parse_args")
    def test_create_cli(self, mock_parse_args):
        """Test if the create_cli function returns the mocked arguments."""
        mock_args = Mock()
        mock_parse_args.return_value = mock_args
        args = create_cli()
        self.assertEqual(args, mock_args)


if __name__ == "__main__":
    unittest.main()
